* Hash6 Tri-en-mmoire-vive.sas : Utiliser une table HASH pour trier une grosse table
                                  SAS en mmoire vive, avec ou sans gestion des doublons;

* Version 1.0 (2013.04.16) Fonctionne en version 9.2 et plus de SAS
          1.1 (2013.04.18) Modifi les listes de variables avec 'ALL:"YES"' au lieu de
              nommer chaque variable
          1.2 (2013.06.03) Corrig un problme dans le tri avec doublons;     

LIBNAME EDUC "C:\EDUCSAS\";
PROC DATASETS  LIBRARY=WORK  NOLIST  KILL;
QUIT;


* A) TRI SUR UNE SEULE CL, SANS GESTION DES DOUBLONS;
/* Procdure SORT classique */
PROC SORT DATA= EDUC.TRANSACTIONS
           OUT= WORK.TRI_S;
  BY ACC;
RUN;

/* quivalent en table HASH */
DATA _NULL_;
    IF 0 THEN SET EDUC.TRANSACTIONS;

    DECLARE HASH TRANS(DATASET:"EDUC.TRANSACTIONS", MULTIDATA:"Y", ORDERED:"Y");
    TRANS.DEFINEKEY("ACC");
    TRANS.DEFINEDATA(ALL:"YES");
    TRANS.DEFINEDONE();

    TRANS.OUTPUT(DATASET:"WORK.TRI_H");
    STOP; /* Un cycle dans l'tape DATA */
RUN;

/* Compare les deux tables en sortie */
PROC COMPARE  BASE=WORK.TRI_S  
              COMPARE=WORK.TRI_H;
  ID ACC;
  FORMAT ACC 10.;
RUN;


* B) TRI SUR UNE SEULE CL, AVEC GESTION DES DOUBLONS;
/* Procdure SORT classique, avec NODUPKEY et table pour les doublons */
PROC SORT DATA  =EDUC.TRANSACTIONS  NODUPKEY
          OUT   =WORK.TRI_S
          DUPOUT=WORK.DOUBLONS_S ;
  BY ACC;
RUN;

/* quivalent en tables HASH - une table HASH pour les premires occurences
   et une autre table HASH pour accueillir les doublons. L'ordre final des 
   observations au sein d`une cl peut tre diffrent de SORT pour un cl donne
   et le choix de laquelle des observations, parmi les doublons d`une cl, se
   trouvera dans chacune des tables de sortie pourrait aussi diffrer */
DATA _NULL_;
    IF 0 THEN SET EDUC.TRANSACTIONS;

    DECLARE HASH CLES_ACC(ORDERED:"Y");
    CLES_ACC.DEFINEKEY("ACC");
    CLES_ACC.DEFINEDATA("ACC");
    CLES_ACC.DEFINEDONE();

    DECLARE HASH PREMIER(ORDERED:"Y");
    PREMIER.DEFINEKEY("ACC");
    PREMIER.DEFINEDATA("ACC", "DEPT", "TX_DATE", "MNT", "QTY");
    PREMIER.DEFINEDONE();

    DECLARE HASH DUP(MULTIDATA:"Y", ORDERED:"Y");
    DUP.DEFINEKEY("ACC");
    DUP.DEFINEDATA("ACC", "DEPT", "TX_DATE", "MNT", "QTY");
    DUP.DEFINEDONE();

    DO UNTIL (FIN);
        SET EDUC.TRANSACTIONS  END=FIN;
        RC= CLES_ACC.FIND(KEY:ACC);
        IF RC NE 0 THEN DO;
            CLES_ACC.ADD();
            PREMIER.ADD();
        END;
        ELSE DUP.ADD();
    END;

    PREMIER.OUTPUT(DATASET:"WORK.TRI_H");
    DUP.OUTPUT(DATASET:"WORK.DOUBLONS_H");
    CLES_ACC.OUTPUT(DATASET:"WORK.CLES_H");
    STOP;
RUN;

/* Compare les deux paires de tables en sortie */
PROC COMPARE  BASE=WORK.TRI_S
              COMP=WORK.TRI_H;
  ID ACC;
  FORMAT ACC 10.;
RUN;
PROC COMPARE  BASE=WORK.DOUBLONS_S
              COMP=WORK.DOUBLONS_H;
  ID ACC;
  FORMAT ACC 10.;
RUN;


* C) TRI SUR DEUX CLS, DONT UNE ASCENDANTE ET L`AUTRE DESCENDANTE;
/* Le fichier de transactions est tri par ordre croissant de compte (ACC)
et en ordre dcroissant de date (TX_DATE). */
PROC SORT DATA= EDUC.TRANSACTIONS
           OUT= TRANS_SORT;
  BY ACC DESCENDING TX_DATE;
RUN;

/* On ne peut mettre une cl croissante et l'autre dcroissante dans la table HASH.
   L'astuce est de placer les rsultats dans une autre table HASH (RESULTS)
   afin de restaurer l'ordre voulu. */
DATA _NULL_;
    IF 0 THEN SET EDUC.TRANSACTIONS;

    DECLARE HASH RESULTS(MULTIDATA:"Y", ORDERED:"Y");
        RESULTS.DEFINEKEY("ACC");
        RESULTS.DEFINEDATA("ACC", "DEPT", "TX_DATE", "MNT", "QTY");
        RESULTS.DEFINEDONE();
    
    DECLARE HASH TRANS(DATASET:"EDUC.TRANSACTIONS", MULTIDATA:"Y", ORDERED:"D");
        TRANS.DEFINEKEY("ACC", "TX_DATE");
        TRANS.DEFINEDATA(ALL:"YES");
        TRANS.DEFINEDONE();
    DECLARE HITER ITER("TRANS");

    RC= ITER.FIRST();
    DO WHILE (RC EQ 0);
        RESULTS.ADD();
        RC= ITER.NEXT();
    END;
    RESULTS.OUTPUT(DATASET:"TRANS_HASH");
RUN;

PROC COMPARE BASE= TRANS_SORT COMPARE= TRANS_HASH;
RUN;

/* Le fichier de transactions est tri par ordre croissant de compte (ACC)
et de DEPT et en ordre dcroissant de date (TX_DATE). */
PROC SORT DATA= EDUC.TRANSACTIONS
           OUT= TRANS_SORT;
  BY ACC DEPT DESCENDING TX_DATE;
RUN;

/* On ne peut mettre une cl croissante et l'autre dcroissante dans la table HASH.
   L'astuce est de placer les rsultats dans une autre table HASH afin de
   restaurer l'ordre voulu. */
DATA _NULL_;
    IF 0 THEN SET EDUC.TRANSACTIONS;

    DECLARE HASH RESULTS(MULTIDATA:"Y", ORDERED:"Y");
        RESULTS.DEFINEKEY("ACC", "DEPT");
        RESULTS.DEFINEDATA("ACC", "DEPT", "TX_DATE", "MNT", "QTY");
        RESULTS.DEFINEDONE();
    
    DECLARE HASH TRANS(DATASET:"EDUC.TRANSACTIONS", MULTIDATA:"Y", ORDERED:"D");
        TRANS.DEFINEKEY("TX_DATE");
        TRANS.DEFINEDATA(ALL:"YES");
        TRANS.DEFINEDONE();
    DECLARE HITER ITER("TRANS");

    RC= ITER.FIRST();
    DO WHILE (RC EQ 0);
        RESULTS.ADD();
        RC= ITER.NEXT();
    END;
    RESULTS.OUTPUT(DATASET:"TRANS_HASH");
RUN;

PROC COMPARE BASE= TRANS_SORT COMPARE= TRANS_HASH;
RUN;
